from ..fileio import makeDirectory, save_pickle

class Experiment():
    """
    A class for running Experiments. 
    The experiment design pattern is based on how lab experiments are performed.
    We have a set of independent test subjects and we record their change over time.


    ...

    Attributes
    ----------
    save_folder : str
        a valid path to a folder for saving the experiment data
    test_subjects : list(TestSubject)
        list of TestSubject objects
    print_report : Boolean
        determines if test output is printed to the terminal

    Methods
    -------
    record()
        Records information from each test subject in test_subjects
    run(epochs=100)
        Runs 100 epochs for each test subject and records the output to the save_folder
    """

    def __init__(self, save_folder, test_subjects, print_report = True):

        self.save_folder = save_folder 
        self.test_subjects = test_subjects
        self.print_report = print_report

    def record(self):

        for test_id, test_subject in enumerate(self.test_subjects):
            test_subject.record(self.save_folder, test_id)
            

    def run(self, epochs=100):

        for epoch in range(epochs):

            if self.print_report : print("Experiment Epoch %d" %(epoch))

            for test_subject in self.test_subjects:
                output = test_subject.measure(epoch = epoch)

                if not output is None and self.print_report:
                    print(output)

            self.record()
        
        #Final Measurements
        for test_subject in self.test_subjects:
            test_subject.final_measure()

        #One last record
        self.record()


from datetime import datetime
from .. import fileio as fio


class TestSubject():
    """
    A class for handling a test subject. 
    The test subject is dependent on some parameters and its output is recorded

    ...

    Attributes
    ----------
    name : str
        name of the test subject

    Methods
    -------
    measure(**kwargs)
        Outputs a measurement dependent on **kwargs
    final_measure(epochs=100)
        Final measurement on the system. If final state has unique behavior, need to let test_subject know that final measurement is occuring.
    record(save_folder, test_id)
        Saves information dictionary to save_folder\test_id
    load(load_folder)
        Loads the information dictionary from load_folder
    """
    __name__ = "TestSubject"

    def __init__(self, name = None):

        self.test_dict = {}
        self.test_dict['Measurements'] = {}
        self.test_dict['Info'] = {}
        
        #Record starttime of test
        self.test_dict['Info']['Absolute Datetime Start'] = fio.datetime_now_str()

        if name is None or type(name) != str:
            self.test_dict['Info']['Name'] = self.__name__
        else:
            self.test_dict['Info']['Name'] = name
            self.__name__ = name
        pass

    def measure(self, **kwargs):
        #Measure state at current epoch
        pass

    def final_measure(self, **kwargs):
        #Final measurement
        self.test_dict['Info']['Absolute Datetime Finish'] = fio.datetime_now_str()

    def record(self, save_folder, test_id):

        test_name = "%s_%d" % (self.__name__, test_id)

        self.test_dict['Info']['Test ID'] = test_name

        test_folder = fio.makeDirectory(save_folder, test_name)

        fio.save_dict_expansion(test_folder, self.test_dict)

    def load(self, load_folder):
        self.test_dict = fio.load_dict_folder(load_folder)

        
        



def get_info(object):
    model_info_op = getattr(object, "info", None)

    if callable(model_info_op):
        return object.info()
    else:
        return "No object.info() function provided"

def dict_str_output(output):

    out_str = ""
    for key in output:
        out_str += "%s : %s\n" %(key, output[key])

    return out_str


from datetime import datetime
import pickle
import numpy as np
import os

def datetime_now_str():

    now = datetime.now()
    return now.strftime("%Y_%m_%d_%H_%M_%S")
    
#Checks to see if a directory exists. If it doesn't it creates it. If it does exist, it does nothing.
def makeDirectory(parent_dir, new_dir):

    save_path = os.path.join(parent_dir, new_dir)

    if not os.path.exists(save_path):
        os.mkdir(save_path)

    return save_path

#Save dictionary as folder structure
def save_dict_expansion(cur_path, dict_struct):

    for key in dict_struct:
        save_type(cur_path, key, dict_struct[key])

def load_dict_folder(cur_path):

    dict_struct = {}

    for f in list(os.scandir(cur_path)):
        
        dict_struct[os.path.splitext(f.name)[0]] = load_type(f.path)

    return dict_struct
        
def load_type(f_path):

    if os.path.isdir(f_path):
        return load_dict_folder(f_path)
    
    extension = os.path.splitext(f_path)[1]

    if extension == '.npy':
        return np.load(f_path)

    elif extension == '.pkl': 
        return load_pickle(f_path)
    
def save_type(cur_path, name, structure):

    if type(structure) == dict and 'is_folder' in structure:
                dict_dir = makeDirectory(cur_path, name)
                save_dict_expansion(dict_dir, structure)
    else: 
        save_pickle(cur_path, name, structure)

def save_dataset(dataset):
    if type(dataset) == np.ndarray:
        np.save()

def load_dataset(data_path):

    split_file = os.path.splitext(data_path)

    if split_file[1] == '.npy':
        return np.load(data_path)

    elif split_file[1] == '.pkl':
        return load_pickle(data_path)

    else:
        return None

def load_pickle(file_str):

    #f = open(file_str, encoding="cp856")
    #p_obj = pickle.load(f)
    #return p_obj
    
    with open(file_str, "rb") as input_file:
        structure = pickle.load(input_file)

    return structure

def save_pickle(folder, file_name, structure):

    file_path = os.path.join(folder, file_name + ".pkl")

    with open(file_path, "wb") as f:
        pickle.dump(structure, f)

def load_test_subject_dicts(folder):
    test_dict = {}
    for subfolder in [ f for f in os.scandir(folder) if f.is_dir() ] :
        key_name = os.path.splitext(subfolder.name)[0]
        test_dict[key_name] = load_dict_folder(subfolder.path)
    return test_dict

#Courtesy of poppie at https://stackoverflow.com/questions/973473/getting-a-list-of-all-subdirectories-in-the-current-directory
def fast_scandir(dirname):
    subfolders= [f.path for f in os.scandir(dirname) if f.is_dir()]
    for dirname in list(subfolders):
        subfolders.extend(fast_scandir(dirname))
    return subfolders